package net;

import game.Game;
import human.HumanDisplay;
import human.HumanModel;
import human.HumanModelData;
import human.HumanStrategy;

import java.awt.BorderLayout;
import java.io.IOException;
import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketException;

import javax.swing.JFrame;
import javax.swing.JOptionPane;

import util.Array;

import lemonade.ScaledLemonadeUtility;


/**
 * This is the class that connects HumanStrategy to a remote server.
 * This class should be robust to any and all server failures. Any
 * illegal message should result in an aborted game, and the client
 * should send a message that the game was aborted.
 * 
 * There are effectively four locks in this system:
 * 1. A lock on the outgoing stream.
 * 2. A lock on the action taken.
 * 3. A lock on the incoming stream.
 * 4. A lock on the model
 * 
 * At present, no thread ever holds more than one lock at a time. 
 * 
 * 1. A lock on the outgoing stream only lasts as long as a message (or urgent message) is being sent.
 * 2. Locks on the action are nearly instantaneous and only involve getting or setting ints and longs
 * 3. A lock on the incoming stream only lasts as long as a message is being received. This can last
 * a while, but no other locks are obtained when the stream is being read.
 * 4. Locks on the model are obtained in one call, then released in a second call: however, no other locks are obtained inside these
 * blocks.
 * 
 * 
 * 
 * For example, the "chat" system does not try to obtain any locks on the model: it has its
 * own display area. 
 * 
 * @author maz
 *
 */
public class HumanRemoteClient implements Runnable {
  Socket s;
  InputStreamListener is;
  OutputStream os;
  HumanStrategy strategy;
  HumanModelData data;
  HumanModel model;
  ChatPanel chatPanel;
  String serverName = "localhost";
  int serverPort = 5000;
  
  int actualPlayerIndex;
  int state;
  Game g;
  
  public static final int INACTIVE=0;
  public static final int ACTIVE=1;
  public static final int SHUTTING_DOWN=2;
  
  
  
  static final int GAMEOVER=0;
  
  static final int PREGAMEABORT=1;
  static final int GAMEABORT=2;
  
  
  /**
   * The current message terminator is CR LF (13 10).
   */
  public static final String messageTerminator = "" + ((char) 13) + ((char) 10);

  public HumanRemoteClient(HumanStrategy strategy, Game g) {
    this.g = g;
    data = new HumanModelData(g.getNumPlayers());
    this.strategy = strategy;
    model = this.strategy.getModel();
  }

  
  public void stop(){
    state = SHUTTING_DOWN;
  }

  /**
   * Send a message to the server. Appends a message terminator (CRLF).
   */
  public synchronized void sendMessage(String message) throws SocketException,
      IOException {
    showVerbose("CLIENT SENDS:" + message);
    message = message + messageTerminator;
    byte[] messageData = message.getBytes();
    if (state==ACTIVE) {
      os.write(messageData);
    }
  }

  /**
   * Send a message to the server. Appends a message terminator (CRLF).
   * This will send a message even if the game is over.
   */
  public synchronized void sendUrgentMessage(String message) throws SocketException,
      IOException {
    showVerbose("CLIENT SENDS:" + message);
    message = message + messageTerminator;
    byte[] messageData = message.getBytes();
    os.write(messageData);
  }

  
  /**
   * Receive a message from the server. Removes a message terminator (period).
   * @throws InterruptedException 
   */
  public String receiveMessage() throws SocketException, IOException, InterruptedException {
    return is.receiveMessage();
  }

  public void closeConnection() throws IOException{
    strategy.setInputStreamListener(null);
    is.close();
    os.close();
    s.close();
  }
  
  public void showVerbose(String string) {
    System.err.println(string);
  }

  /**
   * Test if the message is complete (contains a terminal character)
   */
  public boolean isComplete(String result) {
    return result.endsWith(messageTerminator);
  }

  public void openConnection() throws IOException{
    s = new Socket();
    InetSocketAddress addr = new InetSocketAddress(serverName, serverPort);
    s.connect(addr);
    is = new InputStreamListener(s.getInputStream(),HumanRemoteClient.messageTerminator);
    strategy.setInputStreamListener(is);
    is.addChatListener(this.chatPanel);
    os = s.getOutputStream();
  }

  
  public long getCurrentGameIndex(String gameMessage){
    String[] tokens = gameMessage.split(":| ");
    if (tokens.length<2){
      return 0;
    }
    String numberString = tokens[1];
    try{
      return Long.parseLong(numberString);
    } catch (Exception e){
      return 0;
    }
  }

  
  public String getIdentity(){
    return s.getLocalAddress()+":"+s.getLocalPort();
  }
  /**
   * GAMEOVER if the game ended correctly, PREGAMEABORT if it ended before
   * it began, and GAMEABORT if it was aborted during play.
   * @return type of exit
   * @throws IOException
   */
  public int runGame() throws IOException, InterruptedException{
    strategy.lockAction();
    
    model.lockAndGetModel(data);
    data.setVisibleAll(false);
    data.message = "Trying to connect to "+this.serverName+"...";
    model.setModelAndUnlock(data);
    
    
    try{
      openConnection();
    } catch (IOException IO){
      model.lockAndGetModel(data);
      data.setVisibleAll(false);
      data.message = "Failed to connect to "+this.serverName;
      model.setModelAndUnlock(data);
      throw IO;
    }
    
      
      model.lockAndGetModel(data);
      data.setVisibleAll(false);
      data.message = "Waiting for other players...";
      data.currentGameIndex=getIdentity()+" waiting...";
      
      model.setModelAndUnlock(data);
      String response;

      
      // Receive the "PING" message
      response = receiveMessage();
      if (response.equals("ABORT")){
        return PREGAMEABORT;
      }
      sendMessage("ACK");
      
      
      // Receive the begin lemonade game message
      response = receiveMessage();
      if (response.equals("ABORT")){
        return PREGAMEABORT;
      }
      String[] fields = response.split(" ");
      actualPlayerIndex = Integer.parseInt(fields[fields.length-1]);
      strategy.setActualPlayerIndex(actualPlayerIndex);
      
      model.lockAndGetModel(data);
      data.setBeginGame();
      data.setFaded(0, true);
      data.setVisible(0,false);
      data.setVisible(actualPlayerIndex, true);
      data.setFaded(actualPlayerIndex, false);
      
      data.currentGameIndex=getIdentity()+" "+getCurrentGameIndex(response);
      data.message = "Begin playing!";
      model.setModelAndUnlock(data);

      strategy.unlockAction();
      
      while (true) {
        // This loop plays an action.
        int action = strategy.getAction();
        sendMessage("" + action);
        response = receiveMessage();
        boolean gameOver = response.startsWith("GAMEOVER");
        boolean gameAbort = (response.equals("ABORT"));
        
        if (!gameAbort){
          // NORMAL OUTCOME MESSAGE
          String[] responseParts = response.split(":");

          if (responseParts.length<2){
            return GAMEABORT;
          }
          String[] outcomes = responseParts[1].split(" ");
          int[] intOutcomes = new int[outcomes.length];
          try{
          for (int i = 0; i < outcomes.length; i++) {
            intOutcomes[i] = Integer.parseInt(outcomes[i]);
          }
          } catch (NumberFormatException e){
            return GAMEABORT;
          }
          // If some illegal action is present, just exit.
          if (!g.areLegalActions(intOutcomes)){
            return GAMEABORT;
          }
          
          System.out.println("Outcome:" + Array.arrayToNakedString(intOutcomes));
          strategy.observeOutcome(intOutcomes);
        } else {
          model.lockAndGetModel(data);
          data.setVisibleAll(false);
          model.setModelAndUnlock(data);
        }
        
        
        if (gameOver || gameAbort){
          return GAMEOVER;
        }
      }
  }
  @Override
  public void run()  {
    state = ACTIVE;
    try {
      while (true) {
        int gameEndingType = runGame();
        strategy.lockAction();
        closeConnection();
        if (gameEndingType==PREGAMEABORT){
          continue;
        }
        model.lockAndGetModel(data);
        data.message = "Game Over!";
        data.setFadedAll(true);
        data.utilitiesVisible = false;
        model.setModelAndUnlock(data);
        
        
        
        String status = (gameEndingType==GAMEOVER) ? "Game Over!" : "Player Quit!";
        Object[] options = new Object[] {"Play Again","Quit"};
        Integer result = JOptionPane.showOptionDialog(strategy.getComponent(),
            status, "Play Again?", JOptionPane.DEFAULT_OPTION,
            JOptionPane.QUESTION_MESSAGE, null, options, options[0]);
        System.err.println("Option selected:"+result);
        
        if (result == 1) {
          break;
        }
      }
    } catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (InterruptedException ie){
      System.err.println("HumanRemoteClient thread interrupted: trying to exit cleanly");
      try{
        sendUrgentMessage("12");
      closeConnection();
      } catch (IOException io){
        io.printStackTrace();
      }
      ie.printStackTrace();
    }
    state = INACTIVE;
  }

  public void setChatPanel(ChatPanel panel){
    this.chatPanel = panel;
  }
  public static void main(String[] args) {
    
    
    
    Game g = new Game(new int[] { 12, 12, 12 }, new ScaledLemonadeUtility(12));
    HumanDisplay hdisplay = new HumanDisplay(g);
    HumanStrategy constantStrategy = new HumanStrategy(g, hdisplay);
    HumanRemoteClient client = new HumanRemoteClient(constantStrategy,g);
    ChatPanel panel = new ChatPanel(client);
    
    hdisplay.addMouseListener(constantStrategy);
    JFrame frame = new JFrame("Lemonade Stand Game");
    
    frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    frame.setLayout(new BorderLayout());
    frame.add(hdisplay,BorderLayout.CENTER);
    frame.add(panel,BorderLayout.SOUTH);
    frame.pack();
    frame.validate();
    frame.setLocationRelativeTo(null);
    frame.setVisible(true);
    
    
    
    if (args.length>0){
      client.serverName=args[0];
    }

    client.run();
    constantStrategy.dispose();
    System.exit(0);
  }

}
